unit FlexGif;

interface

uses
  Windows, Classes, Graphics, ExtCtrls, Math,
  RxGif, RxTimer, VCLUtils, RXConst,
  FlexBase, FlexControls, FlexProps, FlexUtils;

type
  // Based on TFlexPicture and TRxGifAnimator from GIFCtrl.pas unit (RX Lib)
  TFlexGifPicture = class(TFlexPicture)
  private
   FIsGif: boolean;
   FGifImage: TGIFImage; // equal PictureProp.Graphic if FIsGif true and nil else
   FTimer: TRxTimer;
   FTimerFrameChange: Boolean;
   FCache: TBitmap;
   FCacheIndex: Integer;
   FCacheTransColor: TColor;
   FAnimateProp: TBoolProp;
   FLoopProp: TBoolProp;
   procedure CheckTimer;
   procedure TimerExpired(Sender: TObject);
   function  GetDelayTime(Index: Integer): Cardinal;
  protected
   procedure CreateProperties; override;
   procedure ControlCreate; override;
   procedure ControlDestroy; override;
   function  GetFrameBitmap(Index: Integer; var TransColor: TColor): TBitmap;
   procedure PropChanged(Sender: TObject; Prop: TCustomProp); override;
   procedure PropStored(Sender: TObject; Prop: TCustomProp;
     var IsStored: boolean; const PropName: string = ''); override;
   procedure Paint(Canvas: TCanvas; var PaintRect: TRect); override;
  public
   property  IsGif: boolean read FIsGif;
   property  AnimateProp: TBoolProp read FAnimateProp;
   property  LoopProp: TBoolProp read FLoopProp;
  end;

implementation

{ Maximum delay (10 sec) guarantees that a very long and slow
  GIF does not hang the system }
const
  MaxDelayTime = 10000;
{$IFDEF WIN32}
  MinDelayTime = 50;
{$ELSE}
  MinDelayTime = 1;
{$ENDIF}

procedure RegisterControl;
begin
 RegisterFlexControl(TFlexGifPicture);
end;

// TFlexGifPicture ////////////////////////////////////////////////////////////

procedure TFlexGifPicture.CreateProperties;
begin
 inherited;
 FAnimateProp := TBoolProp.Create(Props, 'Animate');
 FAnimateProp.Style := FAnimateProp.Style + [ psNonVisual ];
 FLoopProp := TBoolProp.Create(Props, 'Loop');
 FLoopProp.Style := FLoopProp.Style + [ psNonVisual ];
end;

procedure TFlexGifPicture.ControlCreate;
begin
 FTimer := TRxTimer.Create(Nil);
 FTimer.Enabled := false;
 FTimer.OnTimer := TimerExpired;
 inherited;
 FLoopProp.Value := true;
 PictureProp.Masked := true;
end;

procedure TFlexGifPicture.ControlDestroy;
begin
 FCache.Free;
 FTimer.Free;
 inherited;
end;

procedure TFlexGifPicture.Paint(Canvas: TCanvas; var PaintRect: TRect);
var Frame: TBitmap;
    TransColor: TColor;
    PrevRgn, ClipRgn: HRGN;
    ClipTransparent: boolean;
    R: TRect;
begin
 if FIsGif then begin
  // Paint necessary frame
  TransColor := PictureProp.MaskColor;
  Frame := GetFrameBitmap(FrameIndexProp.Value, TransColor);
  ClipRgn := 0;
  PrevRgn := 0;
  ClipTransparent := Owner.UseImageClipTransparent or
    (TransparencyProp.Value > 0);
  try
   R := PaintRect;
   if ClipTransparent and Frame.Transparent then begin
    ClipRgn := CreateRectRgnIndirect(R);
    ExcludeBitmapTransparency(Frame, R, ClipRgn);
    PrevRgn := IntersectClipRgn(Canvas, ClipRgn);
   end;
   if (TransColor = clNone) or not PictureProp.Masked then
    Canvas.StretchDraw(PaintRect, Frame)
   else
    StretchBitmapRectTransparent(Canvas, PaintRect.Left, PaintRect.Top,
      WidthOf(PaintRect), HeightOf(PaintRect),
      Bounds(0, 0, Frame.Width, Frame.Height), Frame, TransColor);
  finally
   if ClipTransparent then begin
    SelectClipRgn(Canvas.Handle, PrevRgn);
    DeleteObject(PrevRgn);
    DeleteObject(ClipRgn);
   end;
  end;
 end else
  inherited;
end;

procedure TFlexGifPicture.PropChanged(Sender: TObject; Prop: TCustomProp);
begin
 if (Prop = PictureProp) and (PictureProp.Graphic <> FGifImage) then begin
  // Stop and reset timer since new image loaded
  FTimer.Reset(True);
  // Reset cache bitmap
  FCacheIndex := -1;
  FCache.Free;
  FCache := nil;
  FCacheTransColor := clNone;
  // Check gif
  FIsGif := Assigned(PictureProp.Graphic) and
    (PictureProp.Graphic is TGIFImage) and not PictureProp.Graphic.Empty;
  if FIsGif
   then FGifImage := TGIFImage(PictureProp.Graphic)
   else FGifImage := Nil;
  PictureProp.MaskColor := FGifImage.TransparentColor;
  CheckTimer;
 end else
 if Prop = FAnimateProp then begin
  CheckTimer;
 end else
 if (Prop = FrameIndexProp) and FIsGif then begin
  // Check loop
  if FrameIndexProp.Value >= FGifImage.Count then begin
   if FLoopProp.Value
    then FrameIndexProp.Value := 0
    else FrameIndexProp.Value := FGifImage.Count - 1;
   exit;
  end;
  // Check frame changing by timer
  if not FTimerFrameChange then begin
   // Direct frame change. Need to reset timer.
   FTimer.Reset(False);
   // Update timer
   CheckTimer;
  end;
 end;
 inherited;
end;

procedure TFlexGifPicture.PropStored(Sender: TObject; Prop: TCustomProp;
  var IsStored: boolean; const PropName: string);
begin
 if Prop = FAnimateProp then
  IsStored := FAnimateProp.Value
 else
 if Prop = FLoopProp then
  IsStored := not FLoopProp.Value
 else
 if (Prop = PictureProp) and (PropName = 'Masked') then
  IsStored := PictureProp.Masked // default true
 else
  inherited;
end;

function TFlexGifPicture.GetFrameBitmap(Index: Integer;
  var TransColor: TColor): TBitmap;
var
  I, Last, First: Integer;
  SavePal: HPalette;
  UseCache: Boolean;
begin
 if not FIsGif then begin
  Result := Nil;
  exit;
 end;
 Index := Min(Index, FGifImage.Count - 1);
 UseCache := (FCache <> nil) and
   (FCacheIndex = Index - 1) and (FCacheIndex >= 0) and
   (FGifImage.Frames[FCacheIndex].DisposalMethod <> dmRestorePrevious);
 if UseCache then begin
  Result := FCache;
  TransColor := FCacheTransColor;
 end else begin
  FCache.Free;
  FCache := nil;
  Result := TBitmap.Create;
 end;
 Result.Canvas.Lock;
 with Result do
 try
  if not UseCache then begin
   Width := FGifImage.ScreenWidth;
   Height := FGifImage.ScreenHeight;
  end;
  Last := Index;
  First := Max(0, Last);
  SavePal := 0;
  if FGifImage.Palette <> 0 then begin
   SavePal := SelectPalette(Canvas.Handle, FGifImage.Palette, False);
   RealizePalette(Canvas.Handle);
  end;
  if not UseCache then begin
   if (FGifImage.Frames[FGifImage.FrameIndex].TransparentColor <> clNone) then begin
    TransColor := GetNearestColor(Canvas.Handle,
      ColorToRGB(FGifImage.Frames[FGifImage.FrameIndex].TransparentColor));
    Canvas.Brush.Color := PaletteColor(TransColor);
   end else
   if (FGifImage.BackgroundColor <> clNone) and FGifImage.Transparent then
    Canvas.Brush.Color := PaletteColor(FGifImage.BackgroundColor)
   else
    Canvas.Brush.Color := PaletteColor(clWindow);
   Canvas.FillRect(Bounds(0, 0, Width, Height));
   while First > 0 do begin
    if (FGifImage.ScreenWidth = FGifImage.Frames[First].Width) and
       (FGifImage.ScreenHeight = FGifImage.Frames[First].Height) then begin
     if (FGifImage.Frames[First].TransparentColor = clNone) or
       ((FGifImage.Frames[First].DisposalMethod = dmRestoreBackground) and
        (First < Last)) then Break;
    end;
    Dec(First);
   end;
   for I := First to Last - 1 do with FGifImage.Frames[I] do
    case DisposalMethod of
     dmUndefined,
     dmLeave:
       Draw(Canvas, Bounds(Origin.X, Origin.Y, Width, Height), True);
     dmRestoreBackground:
       if I > First then
        Canvas.FillRect(Bounds(Origin.X, Origin.Y, Width, Height));
     dmRestorePrevious:
       ; { do nothing }
    end;
  end else
  { UseCache }
  with FGifImage.Frames[FCacheIndex] do
   if DisposalMethod = dmRestoreBackground then
    Canvas.FillRect(Bounds(Origin.X, Origin.Y, Width, Height));
  with FGifImage.Frames[Last] do
   Draw(Canvas, Bounds(Origin.X, Origin.Y, Width, Height), True);
  if (not UseCache) and (TransColor <> clNone) and PictureProp.Masked then begin
   TransparentColor := PaletteColor(TransColor);
   Transparent := True;
  end;
  if FGifImage.Palette <> 0 then SelectPalette(Canvas.Handle, SavePal, False);
  FCache := Result;
  FCacheIndex := Index;
  FCacheTransColor := TransColor;
  Result.Canvas.Unlock;
 except
  Result.Canvas.Unlock;
  if not UseCache then Result.Free;
  raise;
 end;
end;

function TFlexGifPicture.GetDelayTime(Index: Integer): Cardinal;
begin
 Result := 0;
 if not FIsGif or (FGifImage.Count <= 1) or
   (FrameIndexProp.Value < 0) or
   (FrameIndexProp.Value >= FGifImage.Count) then exit;
 Result := FGifImage.Frames[FrameIndexProp.Value].AnimateInterval;
 if Result < MinDelayTime then Result := MinDelayTime else
 if Result > MaxDelayTime then Result := MaxDelayTime;
end;

procedure TFlexGifPicture.CheckTimer;
begin
 if FAnimateProp.Value and FIsGif and
    // Animation looped or it's not last frame
    (FLoopProp.Value or (FrameIndexProp.Value < FGifImage.Count - 1)) then begin
  if (FrameIndexProp.Value >= 0) and (FGifImage.Count > 0) then
   FTimer.Interval := GetDelayTime(FrameIndexProp.Value);
  FTimer.Enabled := True;
 end else begin
  FTimer.Enabled := False;
 end;
end;

procedure TFlexGifPicture.TimerExpired(Sender: TObject);
begin
 FTimerFrameChange := True;
 try
  // Change frame
  if FLoopProp.Value and (FrameIndexProp.Value = FGifImage.Count - 1)
   then FrameIndexProp.Value := 0
   else FrameIndexProp.Value := FrameIndexProp.Value + 1;
 finally
  FTimerFrameChange := False;
 end;
 // Update timer
 CheckTimer;
end;

initialization
  RegisterControl;

end.
