unit ExtFlexControls;

// To use this flex-controls in FlexEdit
// just add this unit in FlexEdit project and
// recompile it.

interface

uses
  Windows, Classes, Graphics, Forms, Controls, StdCtrls,
  FlexBase, FlexProps, FlexUtils, FlexPath;

const
  crCreateTriangleCursor = crLastFlexCursor + 1;

type
  TFlexTriangle = class(TFlexControl)
  private
   FBrushProp: TBrushProp;
   FPenProp: TPenProp;
   FDirectionProp: TIntProp;
  protected
   procedure CreateProperties; override;
   procedure ControlCreate; override;
   procedure ControlTranslate(const TranslateInfo: TTranslateInfo); override;
   function  CreateCurveControl: TFlexControl; override;
   procedure Paint(Canvas: TCanvas; var PaintRect: TRect); override;
   function  GetAnchorPoint: TPoint; override;
   procedure PropChanged(Sender: TObject; Prop: TCustomProp); override;
   procedure PropStored(Sender: TObject; Prop: TCustomProp;
     var IsStored: boolean; const PropName: string = ''); override;
   function  CreateTriangleRegion(const PaintRect: TRect; Inflate: boolean = false): HRGN;
   function  GetRefreshRect(RefreshX, RefreshY: integer): TRect; override;
   function  GetDefaultLinkPoint(Index: integer): TPoint; override;
   function  GetDefaultLinkPointCount: integer; override;
  public
   class function GetToolInfo(ToolIcon: TBitmap; var Hint: string): boolean;
     override;
   class function CursorInCreate: TCursor; override;
   function  IsPointInside(PaintX, PaintY: integer): boolean; override;
   property  BrushProp: TBrushProp read FBrushProp;
   property  PenProp: TPenProp read FPenProp;
   property  DirectionProp: TIntProp read FDirectionProp;
  end;

implementation

{$R Triangle.res}

uses
  FlexControls;

//const
//  SelectionThreshold = 2;

var
  FFlexTriangleToolIcon: TBitmap;

procedure LoadExtResources;
var Cursor: HCursor;
begin
 try
  // Load cursors
  Cursor :=  LoadCursor(HInstance, 'CREATE_TRIANGLE_CUR');
  if Cursor = 0 then begin
   LoadFlexCursors;
   Screen.Cursors[crCreateTriangleCursor] :=
    Screen.Cursors[crCreateControlCursor];
  end else
   Screen.Cursors[crCreateTriangleCursor] := Cursor;
  // Load icons
  if not Assigned(FFlexTriangleToolIcon) then
   FFlexTriangleToolIcon := TBitmap.Create;
  FFlexTriangleToolIcon.LoadFromResourceName(HInstance, 'TRIANGLE_TOOL_ICON');
 except
  Application.HandleException(Nil);
 end;
end;

procedure FreeExtResources;
begin
 FFlexTriangleToolIcon.Free;
 FFlexTriangleToolIcon := Nil;
end;

procedure CreateTriangle(const R: TRect; Points: PPoint; Dir: integer);
const
  Triangles: array[0..3, 0..2] of TPoint = (
    ( (x:0; y:2), (x:1; y:0), (x:2; y:2) ),  // 0 - 90
    ( (x:0; y:0), (x:2; y:1), (x:0; y:2) ),  // 1 - 0
    ( (x:0; y:0), (x:1; y:2), (x:2; y:0) ),  // 2 - 270
    ( (x:2; y:0), (x:0; y:1), (x:2; y:2) )   // 3 - 180
  );
var
  i: integer;
begin
 if (Dir < 0) or (Dir > 3) then exit;
 for i:=0 to 2 do with Triangles[Dir, i] do begin
  Points.x := R.Left + x * (R.Right - R.Left ) div 2;
  Points.y := R.Top  + y * (R.Bottom - R.Top ) div 2;
  inc(Points); 
 end;
end;

// TFlexTriangle //////////////////////////////////////////////////////////////

procedure TFlexTriangle.ControlCreate;
begin
 Width := 1;
 Height := 1;
 FBrushProp.Color := clNone;
 FBrushProp.Style := bsSolid;
 inherited;
 Visible := True;
end;

procedure TFlexTriangle.CreateProperties;
begin
 inherited;
 FBrushProp := TBrushProp.Create(Props, 'Brush');
 FPenProp := TPenProp.Create(Props, 'Pen');
 FDirectionProp := TIntProp.Create(Props, 'Direction');
end;

class function TFlexTriangle.CursorInCreate: TCursor;
begin
 Result := crCreateTriangleCursor;
end;

class function TFlexTriangle.GetToolInfo(ToolIcon: TBitmap;
  var Hint: string): boolean;
begin
 Result := true;
 Hint := 'Triangle tool';
 if Assigned(ToolIcon) and Assigned(FFlexTriangleToolIcon) then
  ToolIcon.Canvas.Draw(0, 0, FFlexTriangleToolIcon);
end;

procedure TFlexTriangle.ControlTranslate(const TranslateInfo: TTranslateInfo);
var Degree: integer;
    Dir: integer;
begin
 inherited;
 Dir := FDirectionProp.Value;
 if TranslateInfo.Mirror and ((Dir = 1) or (Dir = 3)) then Dir := 4 - Dir;
 Degree := (90 - (90 - 90*Dir + TranslateInfo.Rotate) ) mod 360;
 if Degree < 0 then Degree := 360 + Degree;
 FDirectionProp.Value := Degree div 90;
 FBrushProp.Translate(TranslateInfo);
end;

function TFlexTriangle.GetAnchorPoint: TPoint;
var Points: array[0..2] of TPoint;
begin
 CreateTriangle(PaintRect, @Points[0], FDirectionProp.Value);
 Result := Points[1];
end;

function TFlexTriangle.CreateTriangleRegion(const PaintRect: TRect; Inflate: boolean = false): HRGN;
var R: TRect;
    PenWidth: integer;
    Points: array[0..2] of TPoint;
    Rgn: HRGN;
begin
 if (FDirectionProp.Value < 0) or (FDirectionProp.Value > 3) then begin
  Result := 0;
  exit;
 end;
 R := PaintRect;
 PenWidth := ScaleValue(FPenProp.ActiveWidth, Owner.Scale);
 if Inflate then begin
  if (PenWidth > SelectionThreshold) then PenWidth := SelectionThreshold;
  InflateRect(R, PenWidth, PenWidth);
 end;
 // Create trinagle region
 CreateTriangle(R, @Points[0], FDirectionProp.Value);
 Result := CreatePolygonRgn(Points[0], Length(Points), WINDING);
 if Inflate and FBrushProp.IsClear and
   not (Assigned(Owner) and Owner.SelectAsFilled) then begin
  // Restore paint rect
  InflateRect(R, -PenWidth, -PenWidth);
  // Erase triangle "hole"
  PenWidth := ScaleValue(FPenProp.ActiveWidth, Owner.Scale);
  if PenWidth < SelectionThreshold+1 then PenWidth := SelectionThreshold+1;
  InflateRect(R, -PenWidth, -PenWidth);
  // Create trinagle region
  CreateTriangle(R, @Points[0], FDirectionProp.Value);
  Rgn := CreatePolygonRgn(Points[0], Length(Points), WINDING);
  // Create "hole"
  CombineRgn(Result, Result, Rgn, RGN_XOR);
  DeleteObject(Rgn);
 end;
end;

function TFlexTriangle.IsPointInside(PaintX, PaintY: integer): boolean;
var Rgn: HRGN;
begin
 Result := inherited IsPointInside(PaintX, PaintY);
 if not Result then exit;
 // Create trinagle region
 Rgn := CreateTriangleRegion(PaintRect, True);
 Result := PtInRegion(Rgn, PaintX, PaintY);
 DeleteObject(Rgn);
end;

function TFlexTriangle.GetRefreshRect(RefreshX, RefreshY: integer): TRect;
begin
 Result := inherited GetRefreshRect(RefreshX, RefreshY);
 // Check inflate refresh area to pen width
 if Assigned(FPenProp) and (FPenProp.ActiveWidth > 0) then
  InflateRect(Result,
    FPenProp.ActiveWidth div 2 + 2*PixelScaleFactor,
    FPenProp.ActiveWidth div 2 + 2*PixelScaleFactor );
end;

procedure TFlexTriangle.Paint(Canvas: TCanvas; var PaintRect: TRect);
var PrevRgn, ClipRgn: HRGN;
    Points: array of TPoint;

 procedure CanvasSetup;
 begin
  FPenProp.Setup(Canvas, Owner.Scale);
  FBrushProp.Setup(Canvas, Owner.Scale);
 end;

begin
 with Canvas do begin
  CanvasSetup;
  if FBrushProp.IsPaintAlternate then begin
   // Alternate filling triangle region
   PrevRgn := 0;
   ClipRgn := CreateTriangleRegion(PaintRect);
   try
    PrevRgn := IntersectClipRgn(Canvas, ClipRgn);
    FBrushProp.PaintAlternate(Canvas, PaintRect, Owner.PaintRect);
   finally
    SelectClipRgn(Canvas.Handle, PrevRgn);
    DeleteObject(ClipRgn);
    DeleteObject(PrevRgn);
   end;
   CanvasSetup;
  end;
  // Draw triangle as polygon
  SetLength(Points, 3);
  CreateTriangle(PaintRect, @Points[0], FDirectionProp.Value);
  Polygon(Points);
 end;
end;

procedure TFlexTriangle.PropChanged(Sender: TObject; Prop: TCustomProp);
begin
 if Prop = FDirectionProp then begin
  // Check min/max values
  if FDirectionProp.Value < 0 then FDirectionProp.Value := 0 else
  if FDirectionProp.Value > 3 then FDirectionProp.Value := 3;
 end;
 inherited;
end;

procedure TFlexTriangle.PropStored(Sender: TObject; Prop: TCustomProp;
  var IsStored: boolean; const PropName: string = '');
begin
 if Prop = FDirectionProp then
  IsStored := FDirectionProp.Value <> 0
 else
  inherited;
end;

function TFlexTriangle.CreateCurveControl: TFlexControl;
var APoints: TPointArray;
begin
 APoints := Nil;
 // Create triangle curve control
 Result := TFlexCurve.Create(Owner, Parent, Layer);
 try
  Result.BeginUpdate;
  try
   // Copy properties
   FlexControlCopy(Self, Result);
   // Create tiangle points
   SetLength(APoints, 3);
   CreateTriangle(Rect(0, 0, WidthProp.Value, HeightProp.Value), @APoints[0],
     FDirectionProp.Value);
   // Set points
   with TFlexCurve(Result) do begin
    SetPoints(APoints);
    IsSolidProp.Value := true;
   end;
  finally
   Result.EndUpdate;
  end;
 except
  Result.Free;
  raise;
 end;
end;

function TFlexTriangle.GetDefaultLinkPoint(Index: integer): TPoint;
var Points: array[0..2] of TPoint;
begin
 // Check index
 if (Index < Low(Points)) or (Index > High(Points)) then begin
  Result := Point(0, 0);
  exit;
 end;
 // Get point
 CreateTriangle(Rect(0, 0, WidthProp.Value, HeightProp.Value), @Points,
   FDirectionProp.Value);
 Result := Points[Index];
end;

function TFlexTriangle.GetDefaultLinkPointCount: integer;
begin
 // Always 3 link point
 Result := 3;
end;

initialization
  // Load resources for additional flex controls
  LoadExtResources;
  // Register additional flex controls
  RegisterFlexControl(TFlexTriangle);

finalization
  // Free additional flex controls resources
  FreeExtResources;

end.
